import pickle as pickle
from . import http_date
import time
import random
from aquests.lib import pathtool
from aquests.lib import udict
import os, sys
import md5
import urllib.request, urllib.parse, urllib.error
		
class Session:
	def __init__ (self, path, timeout, bsid):
		self.__path = path
		self.__timeout = timeout
		self.__bsid = bsid
		self.__created_time = time.time ()
		self.__last_used = time.time ()
		self.data = data
	
	def __check_timeout (self):
		if time.time () - self.__last_used > self.__timeout:
			self.truncate ()
			raise KeyError("session variable had been timeouted")
			
	def __setitem__ (self, k, v):
		self.data [k] = v
		self.__last_used = time.time ()
		
	def __getitem__ (self, k, v):
		self.__check_timeout ()
		return self.data.get (k, v)
	
	def __delitem__ (self, k, v):
		del self.data [k]
		
	def get_bsid (self):
		return self.__bsid
	
	def get (self, name, default):
		self.__check_timeout ()
		self.__last_used = time.time ()	
		return self.data.get (name, default)
	

class SessionStorage:
	def __init__ (self, path):
		self.path = path
		self.check_directory ()
	
	def check_directory (self):
		if not self.path: return
		for h in list(range(48,58)) + list(range(97,103)):
			for i in list(range(48,58)) + list(range(97,103)):
				for j in list(range(48,58)) + list(range(97,103)):
					initial = "0" + chr (h) + "/" + chr (i) + chr (j)
					pathtool.mkdir (os.path.join (self.path, initial))
	
	def get_filename (self, session):
		if type (session) == type (""):
			bsid = session
		else:	
			bsid = session.get_bsid ()
		return os.path.join (self.path, "0" + bsid[0], bsid[1:3], bsid)
											
	def save (self, session):
		if not self.path: return
		fn = self.get_filename (session)		
		f = open (fn, "wb")
		pickle.dump (session, f)
		f.close ()
		
	def remove (self, session):
		if not self.path: return
		fn = self.get_filename (session)		
		try: os.remove (fn)
		except (OSError): pass	
			
	def load (self, bsid):
		if not self.path: return
		fn = self.get_filename (bsid)		
		f = open (fn, "r")
		session = pickle.load (f)
		f.close ()
		return session
		
	
class Sessions:
	def __init__ (self, path, timeout):
		self.path = path
		self.session_storage = SessionStorage (path)
		self.timeout = timeout
		self.req_count = 0
		self.sessions = {}
		self.load_sessions (self.path)
	
	def make_bsid (self, cookie):
		return md5.new ("%s %s %s" % (
			cookie.request.get_header ("user-agent"), 
			time.time (), 
			random.randrange (1000000)
			)).hexdigest ()
		
	def maintern (self):
		for bsid in list(self.sessions.keys ()):
			session = self.sessions [bsid]
			if time.time () - session.last_used > self.timeout:
				self.remove (bsid)
		
	def get (self, cookie):
		self.req_count += 1
		if self.req_count % 100 == 0:
			self.maintern ()						
		bsid = cookie.get ("BSID")
		if bsid is None:
			bsid = self.make_bsid (cookie)
			cookie.set ("BSID", bsid, "never")
			
		if bsid in self.sessions:			
			session = self.sessions [bsid] 
			if time.time () - session.last_used > self.timeout:
				self.remove (bsid)
		
		if bsid not in self.sessions:
			self.sessions [bsid] = Session (self.path, self.timeout, bsid)
		
		session = self.sessions [bsid]
		return session

	def remove (self, bsid):
		if bsid in self.sessions:
			session = self.sessions [bsid] 
			self.session_storage.remove (session)
			del session
			del self.sessions [bsid]
	
	def load_sessions (self, path):
		for obj in os.listdir (path):
			fn = os.path.join (path, obj)
			if os.path.isdir (fn):
				self.load_sessions (fn)
			else:
				session = self.session_storage.load (fn)	
				self.sessions [session.get_bsid ()] = session
		
	def save_sessions (self):
		for k in self.sessions:
			session = self.sessions [k]
			self.session_storage.save (session)
		
	def cleanup (self):
		self.maintern ()
		self.save_sessions ()
		self.sessions = {}


if __name__ == "__main__":
	def set (name, val = "", expires=None, path = "/", domain = None):
		# browser string cookie
		print(val.serialize ())
		
	
	sess = SecuredCookieValue ({}, "ASDF34x5=DFu$3FD45i&*YTnU+7", set, True)
	sess [3] = (3,4)
	sess.save ()
	
	sess = SecuredCookieValue.unserialize ("zKHqC0mZfX7qyGzVcGGyu3VklHQ=?a=STEKLg==&b=STIKLg==&c=KGxwMQpJMwphSTQKYS4=", "ASDF34x5=DFu$3FD45i&*YTnU+7", set)
	print(sess)
	
	sess = SecuredCookieValue.unserialize ("K05TrqlZvmHtcL461KbiYR+cvjM=?lalala=KEkzCkk0CnRwMQou", "ASDF34x5=DFu$3FD45i&*YTnU+7", set)
	print(sess)
	
	



from .. import collectors
from . import POST_MAX_SIZE


class Collector (collectors.FormCollector):
	asyncon = None
	def __init__ (self, handler, request):
		self.handler = handler
		self.request = request
		self.data = []
		self.cache = []
		self.cached = False
		self.got_all_data = False
		self.length = 0 
		self.content_length = self.get_content_length ()
	
	def reuse_cache (self):
		self.data = self.cache + self.data
		self.cache = []
			
	def start_collect (self):	
		if self.content_length == 0:
			return self.found_terminator ()			
		if self.content_length <= POST_MAX_SIZE: #5M
			self.cached = True
		self.request.channel.set_terminator (self.content_length)
		
	def close (self):
		# channel disconnected
		self.data = []
		self.cache = []
		self.request.collector = None
		
		# abort immediatly
		if self.asyncon:
			self.asyncon.handle_close (710, "Channel Closed")
				
	def collect_incoming_data (self, data):
		self.length += len (data)
		self.data.append (data)

	def found_terminator (self):
		self.request.channel.set_terminator (b'\r\n\r\n')
		self.got_all_data = True
		# don't request.collector = None => do it at callback ()
		# because this collector will be used in Request.continue_start() later
	
	def get_cache (self):
		return b"".join (self.cache)
		
	def ready (self):
		return len (self.data) or self.got_all_data
	
	def exhausted (self):
		return not self.data
	
	def more (self):		
		if not self.data:
			return b""		
		data = b"".join (self.data)
		self.data = []
		if self.cached:			
			self.cache += data		
		return data
		


import multiprocessing

class ModelKeys:
	def __init__ (self, keys):
		self._keys  = keys
		self._arr = multiprocessing.Array ('d', [0.0] * len (keys), lock = multiprocessing.RLock ())		
		self._d = {}
		for i in range (len (keys)):
			self._d [keys [i]] = i
	
	def __len__ (self):
		return len (self.__d)
		
	def __contains__ (self, k):
		with self.__lock:
			return k in self.__d
			
	def __setitem__ (self, k, v):
		self.set (k, v)
	
	def __getitem__ (self, k):
		return self._d [k]
		
	def set (self, k, v, ignore_nokey = False):
		try:
			self._arr [self._d [k]] = v
		except KeyError:
			if not ignore_nokey:
				raise
	
	def get (self, k, d = None):
		try:
			v = self._arr [self._d [k]]
		except KeyError:
			return d
		return v or d
	
	def keys (self):
		return self._keys [:]
	
	def values (self):
		t = []
		for i in range (len (self._keys)):
			t.append (self._arr [i])
		return t
			
	def items (self):
		t = []
		for i in range (len (self._keys)):
			t.append ((self._keys [i], self._arr [i]))
		return t
	
	def has_key (self, k):
		return self.__d.has_key (k)



from . import wsgi_executor
try:
	import xmlrpc.client as xmlrpclib
except ImportError:
	import xmlrpclib
import sys, os
import struct
import asynchat
import threading
import copy
from aquests.protocols.grpc.producers import grpc_producer
from aquests.protocols.grpc.discover import find_input
from aquests.lib.athreads import trigger
from skitai.server.handlers import collectors	
from skitai import version_info, was as the_was

class Executor (wsgi_executor.Executor):
	def __init__ (self, env, get_method):	
		wsgi_executor.Executor.__init__ (self, env, get_method)
		self.producer = None
		self.service = None
		self.num_streams = 0
				
	def __call__ (self):
		request = self.env ["skitai.was"].request
		collector = request.collector
		data = self.env ["wsgi.input"]
		self.input_type = find_input (request.uri [1:])
			
		servicefullname = self.env ["SCRIPT_NAME"][1:-1]
		methodname = self.env ["PATH_INFO"]		
		sfn = servicefullname. split (".")		
		packagename = ".".join (sfn [:-1])
		servicename = sfn [-1]		
		
		current_app, self.service, param, respcode = self.find_method (request, methodname, True)		
		if respcode:
			return b""
			
		self.build_was ()
		self.was.subapp = current_app
		self.was.response ["grpc-accept-encoding"] = 'identity,gzip'
		self.was.response.set_trailer ("grpc-status", "0")
		self.was.response.set_trailer ("grpc-message", "ok")
		
		descriptor = []
		for m in data:
			f = self.input_type [0]()
			f.ParseFromString (m)
			descriptor.append (f)
		if not self.input_type [1]: # not stream
			descriptor = descriptor [0]

		result = b""
		try:
			result = self.chained_exec (self.service, (descriptor,), {})
			
		except:
			self.was.traceback ()			
			self.was.response.set_trailer ("grpc-status", "2")
			self.was.response.set_trailer ("grpc-message", "internal error")		
			self.rollback ()
			
		else:
			if result:
				self.was.response ["content-type"] = "application/grpc"
			self.commit ()
			result = grpc_producer (result [0], False)
			for k,v in result.get_headers ():
				self.was.response [k] = v
			
		del self.was.subapp
		del self.was.env
		
		return result
				
		